#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014-2016 The Regents of the University of California,
#               Antonino Ingargiola <tritemio@gmail.com>
#
"""
The `loader` module contains functions to load each supported data format.
The loader functions load data from a specific format and
return a new :class:`fretbursts.burstlib.Data()` object containing the data.

This module contains the high-level function to load a data-file and
to return a `Data()` object. The low-level functions that perform the binary
loading and preprocessing can be found in the `dataload` folder.
"""

from __future__ import print_function, absolute_import
from builtins import range, zip

import os
import numpy as np
try:
    import cPickle as pickle
except ImportError:
    import pickle
import tables

from .dataload.multi_ch_reader import load_data_ordered16
from .dataload.smreader import load_sm
from .dataload.spcreader import load_spc
from .dataload.manta_reader import (load_manta_timestamps,
                                    load_xavier_manta_data,
                                    get_timestamps_detectors,
                                    # process_timestamps,
                                    process_store,
                                    load_manta_timestamps_pytables)
from .utils.misc import pprint, deprecate
from .burstlib import Data
from . import loader_legacy
import phconvert as phc


def _is_multich(h5data):
    if 'photon_data' in h5data:
        return False
    elif 'photon_data0' in h5data:
        return True
    else:
        msg = 'Cannot find a photon_data group.'
        raise phc.hdf5.Invalid_PhotonHDF5(msg)

def _append_data_ch(d, name, value):
    if name not in d:
        d.add(**{name: [value]})
    else:
        d[name].append(value)

def _load_from_group(d, group, name, dest_name, multich_field=False,
                     ondisk=False, allow_missing=True):
    if allow_missing and name not in group:
        return

    node_value = group._f_get_child(name)
    if not ondisk:
        node_value = node_value.read()
    if multich_field:
        _append_data_ch(d, dest_name, node_value)
    else:
        d.add(**{dest_name: node_value})

def _append_empy_ch(data):
    # Empty channel, fill it with empty arrays
    ph_times = np.array([], dtype='int64')
    _append_data_ch(data, 'ph_times_m', ph_times)

    a_em = np.array([], dtype=bool)
    _append_data_ch(data, 'A_em', a_em)

def _get_measurement_specs(ph_data):
    if 'measurement_specs' not in ph_data:
        # No measurement specs, we will load timestamps and set them all in a
        # conventional photon stream (acceptor emission)
        meas_type = 'smFRET-1color'
        meas_specs = None
    else:
        assert 'measurement_type' in ph_data.measurement_specs
        meas_specs = ph_data.measurement_specs
        meas_type = meas_specs.measurement_type.read().decode()

    if meas_type not in ['smFRET-1color', 'smFRET',
                         'smFRET-usALEX', 'smFRET-nsALEX']:
        raise NotImplementedError('Measurement type "%s" not supported'
                                  ' by FRETBursts.' % meas_type)
    return meas_type, meas_specs

def _load_photon_data_arrays(data, ph_data, meas_type, ondisk=False):
    assert 'timestamps' in ph_data

    # Build mapping to convert Photon-HDF5 to FRETBursts names
    # fields not mapped use the same name on both Photon-HDF5 and FRETBursts
    mapping = {'timestamps': 'ph_times_m',
               'nanotimes': 'nanotimes', 'particles': 'particles'}
    if 'ALEX' in meas_type:
        mapping = {'timestamps': 'ph_times_t', 'detectors': 'det_t',
                   'nanotimes': 'nanotimes_t', 'particles': 'particles_t'}

    # Load all photon-data arrays
    for name in ph_data._v_leaves:
        dest_name = mapping.get(name, name)
        _load_from_group(data, ph_data, name, dest_name=dest_name,
                         multich_field=True, ondisk=ondisk)

    # Timestamps are always present, and their units are always present too
    data.add(clk_p=ph_data.timestamps_specs.timestamps_unit.read())

def _load_nanotimes_specs(data, ph_data):
    nanot_specs = ph_data.nanotimes_specs
    nanotimes_params = {}
    for name in ['tcspc_unit', 'tcspc_num_bins', 'tcspc_range']:
        value = nanot_specs._f_get_child(name).read()
        nanotimes_params.update(**{name: value})
    if 'user' in nanot_specs:
        for name in ['tau_accept_only', 'tau_donor_only',
                     'tau_fret_donor', 'inverse_fret_rate']:
            if name in nanot_specs.user:
                value = nanot_specs.user._f_get_child(name).read()
                nanotimes_params.update(**{name: value})
    _append_data_ch(data, 'nanotimes_params', nanotimes_params)

def _load_alex_periods_donor_acceptor(data, meas_specs):
    # Both us- and ns-ALEX
    try:
        # Try to load alex period definitions
        D_ON = meas_specs.alex_excitation_period1.read()
        A_ON = meas_specs.alex_excitation_period2.read()
    except tables.NoSuchNodeError:
        # But if it fails it's OK, those fields are optional
        print('WARNING: No alternation defintion found.')
    else:
        _append_data_ch(data, 'D_ON', D_ON)
        _append_data_ch(data, 'A_ON', A_ON)

def _compute_acceptor_emission_mask(data, ich):
    """For non-ALEX measurements."""
    if data.detectors[ich].dtype.itemsize != 1:
        raise NotImplementedError('Detectors dtype must be 1-byte.')
    donor, accept = data._det_donor_accept_multich[ich]

    # From `detectors` compute boolean mask `A_em`
    if accept == 0 or donor == 0:
        # In this case we create the boolean mask in-place
        # using the detectors array
        _append_data_ch(data, 'A_em', data.detectors[ich].view(dtype=bool))
        if accept == 0:
            np.logical_not(data.A_em[ich], out=data.A_em[ich])
    else:
        # Create the boolean mask
        _append_data_ch(data, 'A_em', data.detectors[ich] == accept)

def _add_usALEX_specs(data, meas_specs):
    try:
        offset = meas_specs.alex_offset.read()
    except tables.NoSuchNodeError:
        print('WARNING: No offset found, assuming offset = 0.')
        offset = 0
    data.add(offset=offset)
    data.add(alex_period=meas_specs.alex_period.read())

def _photon_hdf5_1ch(h5data, data, ondisk=False, nch=1, ich=0):
    data.add(nch=nch)
    ph_data_name = '/photon_data' if nch == 1 else '/photon_data%d' % ich

    # Handle the case of missing channel (e.g. dead pixel)
    if ph_data_name not in h5data:
        _append_empy_ch(data)
        return

    # Load photon_data group and measurement_specs (if present)
    ph_data = h5data._f_get_child(ph_data_name)
    meas_type, meas_specs = _get_measurement_specs(ph_data)

    # Load photon_data arrays
    _load_photon_data_arrays(data, ph_data, meas_type=meas_type, ondisk=ondisk)

    # If nanotimes are present load their specs
    if 'nanotimes' in ph_data:
        _load_nanotimes_specs(data, ph_data)

    # Unless 1-color, load donor and acceptor info
    if meas_type != 'smFRET-1color':
        donor = np.asscalar(meas_specs.detectors_specs.spectral_ch1.read())
        accept = np.asscalar(meas_specs.detectors_specs.spectral_ch2.read())
        _append_data_ch(data, 'det_donor_accept', (donor, accept))

    # Load alternation definition both for ns-ALEX and us-ALEX
    if 'ALEX' in meas_type:
        _load_alex_periods_donor_acceptor(data, meas_specs)

    # Here there are all the special-case for each measurement type
    if meas_type == 'smFRET-1color':
        # Non-FRET or unspecified data, assume all photons are "acceptor"
        _append_data_ch(data, 'A_em', slice(None))

    elif meas_type == 'smFRET':
        _compute_acceptor_emission_mask(data, ich)

    elif meas_type == 'smFRET-usALEX':
        _add_usALEX_specs(data, meas_specs)

    elif meas_type == 'smFRET-nsALEX':
        data.add(laser_repetition_rate=meas_specs.laser_repetition_rate.read())

    # Set some `data` flags
    data.add(ALEX='ALEX' in meas_type)
    data.add(lifetime='nanotimes' in ph_data)


def _photon_hdf5_multich(h5data, data, ondisk=True):

    ph_times_dict = phc.hdf5.photon_data_mapping(h5data._v_file)
    nch = np.max(list(ph_times_dict.keys())) + 1

    for ich in range(nch):
        _photon_hdf5_1ch(h5data, data, ondisk=ondisk, nch=nch, ich=ich)


def photon_hdf5(filename, ondisk=False, strict=False):
    """Load a data file saved in Photon-HDF5 format version 0.3 or higher.

    Photon-HDF5 is a format for a wide range of timestamp-based
    single molecule data. For more info please see:

    http://photon-hdf5.org/

    Arguments:
        ondisk (bool): if True do not load in the timestamp arrays, just
            load the array reference. Default False.

    Returns:
        :class:`fretbursts.burstlib.Data` object containing the data.
    """
    assert os.path.isfile(filename), 'File not found.'
    version = phc.hdf5._check_version(filename)
    if version == u'0.2':
        return loader_legacy.hdf5(filename)

    h5file = tables.open_file(filename)
    h5data = h5file.root
    d = Data(fname=filename, data_file=h5data._v_file)

    for grp_name in ['setup', 'sample', 'provenance', 'identity']:
        if grp_name in h5data:
            d.add(**{grp_name:
                     phc.hdf5.dict_from_group(h5data._f_get_child(grp_name))})

    for field_name in ['description', 'acquisition_duration']:
        if field_name in h5data:
            d.add(**{field_name: h5data._f_get_child(field_name).read()})

    if _is_multich(h5data):
        _photon_hdf5_multich(h5data, d, ondisk=ondisk)
    else:
        _photon_hdf5_1ch(h5data, d, ondisk=ondisk)

    return d


##
# Multi-spot loader functions
#
def multispot8(fname, bytes_to_read=-1, swap_D_A=True, leakage=0, gamma=1.):
    """Load a 8-ch multispot file and return a Data() object. Cached version.
    """
    fname_c = fname + '_cache.pickle'
    try:
        var = pickle.load(open(fname_c, 'rb'))
        dx = Data(fname=fname, clk_p=12.5e-9, nch=8, leakage=leakage,
                  gamma=gamma)
        dx.add(ph_times_m=var['ph_times_m'], A_em=var['A_em'], ALEX=False)
        pprint(" - File loaded from cache: %s\n" % fname)
    except IOError:
        dx = multispot8_core(fname, bytes_to_read=bytes_to_read,
                             swap_D_A=swap_D_A, leakage=leakage, gamma=gamma)
        D = {'ph_times_m': dx.ph_times_m, 'A_em': dx.A_em}
        pprint(" - Pickling data ... ")
        pickle.dump(D, open(fname_c, 'wb'), -1)
        pprint("DONE\n")
    return dx

load_multispot8 = deprecate(multispot8, "load_multispot8", "loader.multispot8")

def multispot8_core(fname, bytes_to_read=-1, swap_D_A=True, leakage=0,
                    gamma=1.):
    """Load a 8-ch multispot file and return a Data() object.
    """
    dx = Data(fname=fname, clk_p=12.5e-9, nch=8, leakage=leakage,
              gamma=gamma)
    ph_times_m, A_em, ph_times_det = load_data_ordered16(
        fname=fname, n_bytes_to_read=bytes_to_read, swap_D_A=swap_D_A)
    dx.add(ph_times_m=ph_times_m, A_em=A_em, ALEX=False)
    return dx

def multispot48_simple(fname, leakage=0, gamma=1.,
                       i_start=0, i_stop=None, debug=False):
    """Load a 48-ch multispot file and return a Data() object.
    """
    dx = Data(fname=fname, clk_p=10e-9, nch=48, leakage=leakage, gamma=gamma)
    ph_times_m, big_fifo, ch_fifo = load_manta_timestamps(
        fname, i_start=i_start, i_stop=i_stop, debug=debug)
    A_em = [True] * len(ph_times_m)
    dx.add(ph_times_m=ph_times_m, A_em=A_em, ALEX=False)
    big_fifo_full = np.array([b.any() for b in big_fifo]).any()
    ch_fifo_full = np.array([b.any() for b in ch_fifo]).any()
    if big_fifo_full:
        print('WARNING: Big-FIFO full, flags saved in Data()')
        dx.add(big_fifo=big_fifo)
    if ch_fifo_full:
        print('WARNING: CH-FIFO full, flags saved in Data()')
        dx.add(ch_fifo=ch_fifo)
    return dx

def multispot48(fname, leakage=0, gamma=1., reprocess=False,
                i_start=0, i_stop=None, debug=False):
    """Load a 48-ch multispot file and return a Data() object.
    """
    basename, ext = os.path.splitext(fname)
    fname_h5 = basename + '.hdf5'
    fname_dat = basename + '.dat'

    def load_dat_file():
        pprint(' - Loading DAT file: %s ... ' % fname_dat)
        # Load data from raw file and store it in a HDF5 file
        data = load_xavier_manta_data(fname_dat, i_start=i_start,
                                      i_stop=i_stop, debug=debug)
        pprint('DONE.\n - Extracting timestamps and detectors ... ')
        timestamps, det = get_timestamps_detectors(data, nbits=24)
        pprint('DONE.\n - Processing and storing ... ')
        ph_times_m, big_fifo, ch_fifo = process_store(
            timestamps, det, out_fname=fname_h5, fifo_flag=True, debug=False)
        pprint('DONE.\n')
        return ph_times_m, big_fifo, ch_fifo

    if not (os.path.isfile(fname_dat) or os.path.isfile(fname_h5)):
        raise IOError('Data file "%s" not found' % basename)

    if os.path.exists(fname_h5) and not reprocess:
        # There is a HDF5 file
        try:
            pprint(' - Loading HDF5 file: %s ... ' % fname_h5)
            ph_times_m, big_fifo, ch_fifo = \
                load_manta_timestamps_pytables(fname_h5)
            pprint('DONE.\n')
        except tables.HDF5ExtError:
            pprint('\n  Ops! File may be truncated.\n')
            ph_times_m, big_fifo, ch_fifo = load_dat_file()
    else:
        ph_times_m, big_fifo, ch_fifo = load_dat_file()

    # Current data has only acceptor ch
    A_em = [True] * len(ph_times_m)

    dx = Data(fname=fname, clk_p=10e-9, nch=48, leakage=leakage, gamma=gamma)
    dx.add(ph_times_m=ph_times_m, A_em=A_em, ALEX=False,
           data_file=ph_times_m.data_file, bg_data_file=ph_times_m.data_file)
    big_fifo_full = np.array([b[:].any() for b in big_fifo]).any()
    ch_fifo_full = np.array([b[:].any() for b in ch_fifo]).any()
    if big_fifo_full:
        print('WARNING: Big-FIFO full, flags saved in Data()')
        dx.add(big_fifo=big_fifo)
    if ch_fifo_full:
        print('WARNING: CH-FIFO full, flags saved in Data()')
        dx.add(ch_fifo=ch_fifo)
    return dx


##
# usALEX loader functions
#

# Build masks for the alternating periods
def _select_outer_range(times, period, edges):
    return ((times % period) >= edges[0]) + ((times % period) < edges[1])

def _select_inner_range(times, period, edges):
    return ((times % period) >= edges[0]) * ((times % period) < edges[1])

def _select_range(times, period, edges):
    return _select_inner_range(times, period, edges) if edges[0] < edges[1] \
        else _select_outer_range(times, period, edges)

def usalex(fname, leakage=0, gamma=1., header=None, start=None, stop=None,
           BT=None):
    """Load usALEX data from a SM file and return a Data() object.

    This function returns a Data() object to which you need to apply
    an alternation selection before performing further analysis (background
    estimation, burst search, etc.).

    The pattern to load usALEX data is the following::

        d = loader.usalex(fname=fname)
        d.add(D_ON=(2850, 580), A_ON=(900, 2580), alex_period=4000)
        plot_alternation_hist(d)

    If the plot looks good, apply the alternation with::

        loader.alex_apply_period(d)

    Now `d` is ready for futher processing such as background estimation,
    burst search, etc...
    """
    if BT is not None:
        print('WARNING: `BT` argument is deprecated, use `leakage` instead.')
        leakage = BT
    if header is not None:
        print('WARNING: `header` argument ignored. '
              '         The header length is now computed automatically.')
    print(" - Loading '%s' ... " % fname)
    ph_times_t, det_t, labels = load_sm(fname, start=start, stop=stop,
                                        return_labels=True)
    print(" [DONE]\n")

    DONOR_ON = (2850, 580)
    ACCEPT_ON = (930, 2580)
    alex_period = 4000

    dx = Data(fname=fname, clk_p=12.5e-9, nch=1, leakage=leakage, gamma=gamma,
              ALEX=True, lifetime=False,
              D_ON=DONOR_ON, A_ON=ACCEPT_ON, alex_period=alex_period,
              ph_times_t=[ph_times_t], det_t=[det_t], det_donor_accept=(0, 1),
              ch_labels=labels)
    return dx

def usalex_apply_period(d, delete_ph_t=True, remove_d_em_a_ex=False):
    """Applies to the Data object `d` the alternation period previously set.

    Note that you first need to load the data in a variable `d` and then
    set the alternation parameters using `d.add(D_ON=..., A_ON=...)`.

    The typical pattern for loading ALEX data is the following::

        d = loader.photon_hdf5(fname=fname)
        d.add(D_ON=(2850, 580), A_ON=(900, 2580))
        alex_plot_alternation(d)

    If the plot looks good, apply the alternation with::

        loader.alex_apply_period(d)

    Now `d` is ready for futher processing such as background estimation,
    burst search, etc...

    *See also:* :func:`alex_apply_period`.
    """
    ich = 0
    donor_ch, accept_ch = d._det_donor_accept_multich[ich]
    D_ON, A_ON = d._D_ON_multich[ich], d._A_ON_multich[ich]
    # Remove eventual ch different from donor or acceptor
    d_ch_mask_t = (d.det_t[ich] == donor_ch)
    a_ch_mask_t = (d.det_t[ich] == accept_ch)
    valid_mask = d_ch_mask_t + a_ch_mask_t
    ph_times_val = d.ph_times_t[ich][valid_mask]
    d_ch_mask_val = d_ch_mask_t[valid_mask]
    a_ch_mask_val = a_ch_mask_t[valid_mask]
    assert (d_ch_mask_val + a_ch_mask_val).all()
    assert not (d_ch_mask_val * a_ch_mask_val).any()
    if 'offset' in d:
        ph_times_val -= d.offset

    # Build masks for excitation windows
    d_ex_mask_val = _select_range(ph_times_val, d.alex_period, D_ON)
    a_ex_mask_val = _select_range(ph_times_val, d.alex_period, A_ON)
    # Safety check: each ph is either D or A ex (not both)
    assert not (d_ex_mask_val * a_ex_mask_val).any()

    # Select alternation periods, removing transients
    DexAex_mask = d_ex_mask_val + a_ex_mask_val

    # Reduce timestamps and masks to the DexAex_mask selection
    ph_times = ph_times_val[DexAex_mask]
    d_em = d_ch_mask_val[DexAex_mask]
    a_em = a_ch_mask_val[DexAex_mask]
    d_ex = d_ex_mask_val[DexAex_mask]
    a_ex = a_ex_mask_val[DexAex_mask]
    assert d_ex.sum() == d_ex_mask_val.sum()
    assert a_ex.sum() == a_ex_mask_val.sum()

    if remove_d_em_a_ex:
        # Removes donor-ch photons during acceptor excitation
        mask = a_em + d_em * d_ex
        assert (mask == -(a_ex * d_em)).all()

        ph_times = ph_times[mask]
        d_em = d_em[mask]
        a_em = a_em[mask]
        d_ex = d_ex[mask]
        a_ex = a_ex[mask]

    assert d_em.sum() + a_em.sum() == ph_times.size
    assert (d_em + a_em).all()       # masks fill the total array
    assert not (d_em * a_em).any()   # no photon is both D and A
    assert a_ex.size == a_em.size == d_ex.size == d_em.size == ph_times.size
    d.add(ph_times_m=[ph_times],
          D_em=[d_em], A_em=[a_em], D_ex=[d_ex], A_ex=[a_ex])

    if 'particles_t' in d:
        particles_val = d.particles_t[ich][valid_mask]
        particles = particles_val[DexAex_mask]
        d.add(particles=[particles])

    assert d.ph_times_m[ich].size == d.A_em[ich].size

    if delete_ph_t:
        d.delete('ph_times_t')
        d.delete('det_t')
    return d


##
# nsALEX loader functions
#

def nsalex(fname):
    """Load nsALEX data from a SPC file and return a Data() object.

    This function returns a Data() object to which you need to apply
    an alternation selection before performing further analysis (background
    estimation, burst search, etc.).

    The pattern to load nsALEX data is the following::

        d = loader.nsalex(fname=fname)
        d.add(D_ON=(2850, 580), A_ON=(900, 2580))
        alex_plot_alternation(d)

    If the plot looks good apply the alternation with::

        loader.alex_apply_period(d)

    Now `d` is ready for futher processing such as background estimation,
    burst search, etc...
    """
    ph_times_t, det_t, nanotimes = load_spc(fname)

    DONOR_ON = (10, 1500)
    ACCEPT_ON = (2000, 3500)
    nanotimes_nbins = 4095

    dx = Data(fname=fname, clk_p=50e-9, nch=1, ALEX=True, lifetime=True,
              D_ON=DONOR_ON, A_ON=ACCEPT_ON,
              nanotimes_nbins=nanotimes_nbins,
              nanotimes_params=[{'tcspc_num_bins': nanotimes_nbins}],
              ph_times_t=[ph_times_t], det_t=[det_t], nanotimes_t=[nanotimes],
              det_donor_accept=(4, 6))
    return dx

def nsalex_apply_period(d, delete_ph_t=True):
    """Applies to the Data object `d` the alternation period previously set.

    Note that you first need to load the data in a variable `d` and then
    set the alternation parameters using `d.add(D_ON=..., A_ON=...)`.

    The typical pattern for loading ALEX data is the following::

        d = loader.photon_hdf5(fname=fname)
        d.add(D_ON=(2850, 580), A_ON=(900, 2580))
        alex_plot_alternation(d)

    If the plot looks good, apply the alternation with::

        loader.alex_apply_period(d)

    Now `d` is ready for futher processing such as background estimation,
    burst search, etc...

    *See also:* :func:`alex_apply_period`.
    """
    ich = 0
    donor_ch, accept_ch = d._det_donor_accept_multich[ich]
    D_ON_multi, A_ON_multi = d._D_ON_multich[ich], d._A_ON_multich[ich]
    D_ON = [(D_ON_multi[i], D_ON_multi[i+1]) for i in range(0, len(D_ON_multi), 2)]
    A_ON = [(A_ON_multi[i], A_ON_multi[i+1]) for i in range(0, len(A_ON_multi), 2)]

    # Mask for donor + acceptor detectors (discard other detectors)
    d_ch_mask_t = (d.det_t[ich] == donor_ch)
    a_ch_mask_t = (d.det_t[ich] == accept_ch)
    da_ch_mask_t = d_ch_mask_t + a_ch_mask_t

    # Masks for excitation periods
    d_ex_mask_t = np.zeros(d.nanotimes_t[ich].size, dtype='bool')
    for d_on in D_ON:
        d_ex_mask_t += (d.nanotimes_t[ich] > d_on[0]) * (d.nanotimes_t[ich] < d_on[1])

    a_ex_mask_t = np.zeros(d.nanotimes_t[ich].size, dtype='bool')
    for a_on in A_ON:
        a_ex_mask_t += (d.nanotimes_t[ich] > a_on[0]) * (d.nanotimes_t[ich] < a_on[1])

    ex_mask_t = d_ex_mask_t + a_ex_mask_t  # Select only ph during Dex or Aex

    # Total mask: D+A photons, and only during the excitation periods
    mask = da_ch_mask_t * ex_mask_t  # logical AND

    # Apply selection to timestamps and nanotimes
    ph_times = d.ph_times_t[ich][mask]
    nanotimes = d.nanotimes_t[ich][mask]

    # Apply selection to the emission masks
    d_em = d_ch_mask_t[mask]
    a_em = a_ch_mask_t[mask]
    assert (d_em + a_em).all()       # masks fill the total array
    assert not (d_em * a_em).any()   # no photon is both D and A

    # Apply selection to the excitation masks
    d_ex = d_ex_mask_t[mask]
    a_ex = a_ex_mask_t[mask]
    assert (d_ex + a_ex).all()
    assert not (d_ex * a_ex).any()

    d.add(ph_times_m=[ph_times], nanotimes=[nanotimes],
          D_em=[d_em], A_em=[a_em], D_ex=[d_ex], A_ex=[a_ex],)

    if delete_ph_t:
        d.delete('ph_times_t')
        d.delete('det_t')
        d.delete('nanotimes_t')

def alex_apply_period(d, delete_ph_t=True):
    """Apply the ALEX period definition set in D_ON and A_ON attributes.

    This function works both for us-ALEX and ns-ALEX data.

    Note that you first need to load the data in a variable `d` and then
    set the alternation parameters using `d.add(D_ON=..., A_ON=...)`.

    The typical pattern for loading ALEX data is the following::

        d = loader.photon_hdf5(fname=fname)
        d.add(D_ON=(2850, 580), A_ON=(900, 2580))
        alex_plot_alternation(d)

    If the plot looks good, apply the alternation with::

        loader.alex_apply_period(d)

    Now `d` is ready for futher processing such as background estimation,
    burst search, etc...
    """
    assert d.ALEX
    if d.lifetime:
        apply_period_func = nsalex_apply_period
    else:
        apply_period_func = usalex_apply_period
    apply_period_func(d, delete_ph_t=delete_ph_t)
    msg = ('# Total photons (after ALEX selection):  {:10,}\n'
           '#  D  photons in D+A excitation periods: {:10,}\n'
           '#  A  photons in D+A excitation periods: {:10,}\n'
           '# D+A photons in  D  excitation period:  {:10,}\n'
           '# D+A photons in  A  excitation period:  {:10,}\n')
    print(msg.format(d.ph_times_m[0].size, d.D_em[0].sum(), d.A_em[0].sum(),
                     d.D_ex[0].sum(), d.A_ex[0].sum()))



# encoding: utf-8
#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014 Antonino Ingargiola <tritemio@gmail.com>
#
"""
WARNING: Plot function ATTIC! Functions here are broken!

Here there are function originally placed in burst_plot.py that became
broken and I didn't had the time or the need to update. They live here until
I decide to fix or delete them.
"""

##
#  Other plot wrapper functions
#

def wplot(*args, **kwargs):
    AX, s = dplot_8ch(*args, **kwargs)
    kwargs.update(AX=AX)
    q = gs.mToolQT(gcf(), dplot_8ch, *args, **kwargs)
    return AX, q

def splot(d, fun=scatter_width_size,
        scroll=False, pgrid=True, figsize=(10, 8), nosuptitle=False, ax=None,
        scale=True, **kwargs):
    if ax is None:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)
    else:
        fig = ax.figure
    for i in xrange(d.nch):
        try:
            if i == 0 and not nosuptitle: fig.set_title(d.status())
        except:
            print("WARNING: No title in plots.")
        ax.grid(pgrid)
        fun(d, i, **kwargs)
    s = None
    if scroll: s = ScrollingToolQT(fig)
    return ax, s

##
#  Other misc plot functions
#
def bplot(d, ich, b_index,  ph0=True, pad=0):
    """Plot photons in a burst as vertical bars. Burst: d.mburst[ich][b_index].
    """
    br = bl.b_irange(d.mburst[ich], b_index, pad=pad)
    accept = d.A_em[ich][br]
    donor = -accept
    ph = d.ph_times_m[ich][br]
    if ph0: ph -= ph[0]
    dtime = (ph[donor])*d.clk_p*1e6
    atime = (ph[accept])*d.clk_p*1e6
    plt.vlines(dtime,0,1, lw=2, color='g', alpha=0.8)
    plt.vlines(atime,0,1, lw=2, color='r', alpha=0.8)
    #plot(dtime, ones_like(ph[donor]), '^', color='g', alpha=0.5)
    #plot(atime, -ones_like(ph[accept]), 'o', color='r', alpha=0.5)
    xlabel("Time (us)")
    nd, na, nt = donor.sum(), accept.sum(), ph.size
    E = float(na)/(nt)
    title("#ph = %d, #D-ph = %d, #A-ph = %d, E = %.2f" % (nt,nd,na,E))
    plt.ylim(-10,10)


def bg_legend_8ch(d):
    ax = gca()
    L = ax.get_lines()[1::2]
    for i,l in enumerate(L):
        ich = i/3
        x = i%3
        s = ['Tot', 'D', 'A']
        r = [d.rate_m[ich], d.rate_dd[ich], d.rate_ad[ich]]
        l.set_label("CH%d, %s %d cps" % (ich+1, s[x], r[x]))
    ax.legend()
    plt.draw()

def bg_legend_alex(d):
    ax = gca()
    L = ax.get_lines()[1::2]
    for i,l in enumerate(L):
        ich = i/4
        x = i%4
        s = ['Tot', 'DD', 'AD', 'AA']
        r = [d.rate_m[ich], d.rate_dd[ich], d.rate_ad[ich], d.rate_aa[ich]]
        l.set_label("CH%d, %s %d cps" % (ich+1, s[x], r[x]))
    ax.legend()
    plt.draw()



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014-2016 The Regents of the University of California,
#               Antonino Ingargiola <tritemio@gmail.com>
#
"""
This module contains a function to store :class:`fretbursts.burstlib.Data`
objects to disk in **Photon-HDF5** format.

Utility functions to print the HDF5 file structure and data-attributes are
also provided.
"""

from __future__ import print_function, absolute_import
from builtins import range, zip

import phconvert as phc


hdf5_data_map = dict(
    filename='fname',
    timestamps_unit='clk_p',
    num_spots='nch',
    alex='ALEX',
    #lifetime
    #alex_period
    timestamps='ph_times_t',
    detectors='det_t',
    #nanotimes='nanotime',
    #particles='par',
    alex_period_donor='D_ON',
    alex_period_acceptor='A_ON',
)

hdf5_data_map_r = {v: k for k, v in hdf5_data_map.items()}

def store(d, compression=dict(complevel=6, complib='zlib'), h5_fname=None,
          verbose=True, num_spectral_ch=2):
    """
    Saves the `Data` object `d` in the Photon-HDF5 format.

    As a side effect the `d` object is modified by adding the attribute
    `data_file` that contains a reference to the pytables file.

    Arguments:
        d (Data object): the Data object containing the smFRET measurement.
        compression (dict): a dictionary containing the compression type
            and level. Passed to pytables `tables.Filters()`.
        h5_fname (string or None): if not None, contains the file name
            to be used for the HDF5 file. If None, the file name is generated
            from `d.fname`, by replacing the original extension with '.hdf5'.
        verbose (bool): if True prints the name of the saved file.

    For description and specs of the Photon-HDF5 format see:
    http://photon-hdf5.readthedocs.org/
    """
    print('DEPRECATED: This function saves the Photon-HDF5 0.2 format '
          'which is deprecated. \n            Please use '
          '`save_photon_hdf5()` in `phconvert` to save version >=0.3.')
    #comp_filter = tables.Filters(**compression)
    if 'lifetime' not in d:
        # Test on different fields for ALEX and non-ALEX
        d.add(lifetime=('nanotimes_t' in d) or ('nanotimes' in d))

    # Add default values for missing mandatory fields
    if 'num_spectral_ch' not in d:
        d.add(num_spectral_ch=num_spectral_ch)
    if 'num_polariz_ch' not in d:
        d.add(num_polariz_ch=1)
    if 'num_detectors' not in d:
        d.add(num_detectors=d.nch*d.num_spectral_ch)

    data = {hdf5_data_map_r.get(k, k): v for k, v in d.items()}

    if d.lifetime:
        data.update(d.nanotimes_params)

    if d.nch == 1:
        if d.ALEX:
            data['donor'], data['acceptor'] = d.det_donor_accept
            pass
        else:
            data['timestamps'] = data.pop('ph_times_m')[0]
            data['detectors'] = data.pop('A_em')[0]
            data['donor'], data['acceptor'] = 1, 0
            if 'par' in d:
                data['particles'] = data.pop('par')[0]
        phc.hdf5.photon_hdf5(data, compression=compression)
    else:
        if d.ALEX:
            raise NotImplementedError
        else:
            data['donor'], data['acceptor'] = 1, 0
            det = [None if isinstance(a_em, slice) else a_em.view('uint8')
                   for a_em in d.A_em]
            phc.hdf5.photon_hdf5(data, compression=compression,
                                 iter_timestamps=d.iter_ph_times(),
                                 iter_detectors=iter(det))
    d.add(data_file=data['data_file'])



#from gaussian_fitting import *
import numpy.random as R

mu0 = 0
sigma0 = 1
print("Real values: %s " % (mu0, sigma0))

## Compare the two methods
n = 1000
Mu1, Sig1 = zeros(n), zeros(n)
Mu2, Sig2 = zeros(n), zeros(n)
Mu3, Sig3 = zeros(n), zeros(n)
for i in range(n):
    s = R.normal(size=2000, loc=mu0, scale=sigma0)
    s = s[s>-0.8]
    #print "Empirical mean and std", s.mean(), s.std()
    m1,s1 = GF.gaussian_fit_cdf(s)
    m2,s2 = GF.gaussian_fit_hist(s, bins=10)
    m3,s3 = s.mean(), s.std()

    Mu1[i] = m1; Sig1[i] = s1
    Mu2[i] = m2; Sig2[i] = s2
    Mu3[i] = m3; Sig3[i] = s3

figure(1); title("Mu")
b = r_[mu0-1:mu0+1:40j]
kw = dict(alpha=0.4, histtype='stepfilled')
hist(Mu1, bins=b, label='PDF', **kw)
hist(Mu2, bins=b, label='Histo', **kw)
#hist(Mu3, bins=b, label='Moments', **kw)
legend(loc='best')
axvline(mu0, color='k', lw=2)
grid(True)

figure(2); title("Sigma")
b = r_[sigma0-1:sigma0+1:40j]
kw = dict(alpha=0.4, histtype='stepfilled')
hist(Sig1, bins=b, label='PDF', **kw)
hist(Sig2, bins=b, label='Histo', **kw)
#hist(Sig3, bins=b, label='Moments', **kw)
legend(loc='best')
axvline(sigma0, color='k', lw=2)
grid(True)

show()





from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
import numpy as NP

ext_modules = [Extension("burstsearch_c", ["burstsearch_c.pyx"])]

setup(
  name = 'Burst search',
  cmdclass = {'build_ext': build_ext},
  include_dirs = [NP.get_include()],
  ext_modules = ext_modules
)



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014-2016 Antonino Ingargiola <tritemio@gmail.com>
#
"""
Generic functions to fit exponential populations.

These functions can be used directly, or, in a typical FRETBursts workflow
they are passed to higher level methods.

*See also:*

* :doc:`background`
"""

from builtins import range, zip

import numpy as np
from scipy.stats import linregress, expon
from scipy.optimize import leastsq


def get_ecdf(s, offset=0.5):
    """Return arrays (x, y) for the empirical CDF curve of sample `s`.

    See the code for more info (is a one-liner!).

    Arguments:
        s (array of floats): sample
        offset (float, default 0.5): Offset to add to the y values of the CDF

    Returns:
        (x, y) (tuple of arrays): the x and y values of the empirical CDF
    """
    return np.sort(s), np.arange(offset, s.size+offset)*1./s.size

def get_residuals(s, tau_fit, offset=0.5):
    """Returns residuals of sample `s` CDF vs an exponential CDF.

    Arguments:
        s (array of floats): sample
        tau_fit (float): mean waiting-time of the exponential distribution
            to use as reference
        offset (float): Default 0.5. Offset to add to the empirical CDF.
            See :func:`get_ecdf` for details.

    Returns:
        residuals (array): residuals of empirical CDF compared with analytical
        CDF with time constant `tau_fit`.
    """
    x, y = get_ecdf(s, offset=offset)
    ye = expon.cdf(x, scale=tau_fit)
    residuals = y - ye
    return x, residuals

def expon_fit(s, s_min=0, offset=0.5, calc_residuals=True):
    """Fit sample `s` to an exponential distribution using the ML estimator.

    This function computes the rate (Lambda) using the maximum likelihood (ML)
    estimator of the mean waiting-time (Tau), that for an exponentially
    distributed sample is the sample-mean.

    Arguments:
        s (array): array of exponetially-distributed samples
        s_min (float): all samples < `s_min` are discarded
            (`s_min` must be >= 0).
        offset (float): offset for computing the CDF. See :func:`get_ecdf`.
        calc_residuals (bool): if True compute the residuals of the fitted
            exponential versus the empirical CDF.

    Returns:
        A 4-tuple of the fitted rate (1/life-time), residuals array,
        residuals x-axis array, sample size after threshold.
    """
    if s_min > 0: s = s[s >= s_min] - s_min
    assert s.size > 10

    # Maximum likelihood estimator of the waiting-time
    Lambda = 1./s.mean()

    x_residuals, residuals = None, None
    if calc_residuals:
        x_residuals, residuals = get_residuals(s, tau_fit=1./Lambda,
                                               offset=offset)
    return Lambda, residuals, x_residuals, s.size

def expon_fit_cdf(s, s_min=0, offset=0.5, calc_residuals=True):
    """Fit of an exponential model to the empirical CDF of `s`.

    This function computes the rate (Lambda) fitting a line (linear
    regression) to the log of the empirical CDF.

    Arguments:
        s (array): array of exponetially-distributed samples
        s_min (float): all samples < `s_min` are discarded
            (`s_min` must be >= 0).
        offset (float): offset for computing the CDF. See :func:`get_ecdf`.
        calc_residuals (bool): if True compute the residuals of the fitted
            exponential versus the empirical CDF.

    Returns:
        A 4-tuple of the fitted rate (1/life-time), residuals array,
        residuals x-axis array, sample size after threshold.
    """
    if s_min > 0: s = s[s >= s_min] - s_min
    assert s.size > 10

    # Line fit the log of the eCDF to compute the rate (Lambda)
    ecdf = get_ecdf(s, offset=offset)
    decr_line = np.log(1-ecdf[1])
    L = linregress(ecdf[0], decr_line)
    Lambda = -L[0]

    x_residuals, residuals = None, None
    if calc_residuals:
        x_residuals, residuals = get_residuals(s, tau_fit=1./Lambda,
                                               offset=offset)
    return Lambda, residuals, x_residuals, s.size


def expon_fit_hist(s, bins, s_min=0, weights=None, offset=0.5,
                   calc_residuals=True):
    """Fit of an exponential model to the histogram of `s` using least squares.

    Arguments:
        s (array): array of exponetially-distributed samples
        bins (float or array): if float is the bin width, otherwise is the
            array of bin edges (passed to `numpy.histogram`)
        s_min (float): all samples < `s_min` are discarded
            (`s_min` must be >= 0).
        weights (None or string): if None no weights is applied.
            if is 'hist_counts', each bin has a weight equal to its counts
            if is 'inv_hist_counts', the weight is the inverse of the counts.
        offset (float): offset for computing the CDF. See :func:`get_ecdf`.
        calc_residuals (bool): if True compute the residuals of the fitted
            exponential versus the empirical CDF.

    Returns:
        A 4-tuple of the fitted rate (1/life-time), residuals array,
        residuals x-axis array, sample size after threshold.
    """
    if s_min > 0: s = s[s >= s_min] - s_min
    assert s.size > 10

    counts, bins = np.histogram(s, bins=bins, density=True)
    x = bins[:-1] + 0.5*(bins[1] - bins[0])  # bin center position
    y = counts
    x = x[y > 0]
    y = y[y > 0]

    if weights is None:
        w = np.ones(y.size)
    elif weights == 'hist_counts':
        w = np.sqrt(y*s.size*(bins[1]-bins[0]))
    elif weights == 'inv_hist_counts':
        w = np.sqrt(1./y*s.size*(bins[1]-bins[0]))
    else:
        raise ValueError('Weighting scheme not valid (use: None, or '
                         '"hist_counts")')

    exp_fun = lambda x, rate: rate*np.exp(-x*rate)
    err_fun = lambda rate, x, y, w: (exp_fun(x, rate) - y)*w

    res = leastsq(err_fun, x0=1./(s.mean()), args=(x, y, w))
    Lambda = res[0]

    x_residuals, residuals = None, None
    if calc_residuals:
        x_residuals, residuals = get_residuals(s, tau_fit=1./Lambda,
                                               offset=offset)
    return Lambda, residuals, x_residuals, s.size



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014 Antonino Ingargiola <tritemio@gmail.com>
#
"""
A global plot style for matplotlib.
"""

from matplotlib.pyplot import rcParams


# This will fail on ReadTheDocs so I use a try-except
try:
    fontsize = fs = 12

    font1 = {'fontname':'Liberation Sans','fontsize':16}
    font2 = {'fontname':'Arial','fontsize':16}

    #rcParams["font.family"] = font2['fontname']
    rcParams["font.sans-serif"] = ['Arial', 'Liberation Sans']
    rcParams["font.size"] = fontsize

    rcParams['xtick.labelsize'] = fontsize
    rcParams['ytick.labelsize'] = fontsize
    rcParams['axes.labelsize'] = fontsize
    rcParams['legend.fontsize'] = fontsize - 1
    rcParams['lines.linewidth'] = 2

    # Define 8 different colors (1 per ch)
    rcParams['axes.color_cycle'] = ['b', 'g', 'r', 'c', 'm', 'y', 'k', 'orange']
except:
    pass



# Copyright (c) 2015-2016 Antonino Ingargiola <tritemio@gmail.com>
# License: MIT
"""
Script to execute notebooks in a folder, unless their name
starts with '_' or ends with '-out'.

The executed notebooks are saved with a '-out' suffix.

Usage:

    nbrun.py notebook_folder ouput_folder

"""
from __future__ import print_function
import os
import time
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor


def run_notebook(notebook_name, nb_suffix='-out', out_path='.', timeout=3600,
                 **execute_kwargs):
    """Runs a notebook and saves the output in the same notebook.

    Arguments:
        notebook_name (string): name of the notebook to be executed.
        nb_suffix (string): suffix to append to the file name of the executed
            notebook.
        timeout (int): timeout in seconds after which the execution is aborted.
        execute_kwargs (dict): additional arguments passed to
            `ExecutePreprocessor`.
        out_path (string): folder where to save the output notebook.
    """
    timestamp_cell = "**Executed:** %s\n\n**Duration:** %d seconds."

    if str(notebook_name).endswith('.ipynb'):
        notebook_name = str(notebook_name)[:-len('.ipynb')]
    nb_name_input = notebook_name + '.ipynb'
    nb_name_output = notebook_name + '%s.ipynb' % nb_suffix
    nb_name_output = os.path.join(out_path, nb_name_output)
    print('- Executing: ', nb_name_input)

    execute_kwargs_ = dict(kernel_name = 'python%d' % sys.version_info[0])
    if execute_kwargs is not None:
        execute_kwargs_.update(execute_kwargs)
    ep = ExecutePreprocessor(timeout=timeout, **execute_kwargs_)
    nb = nbformat.read(nb_name_input, as_version=4)

    start_time = time.time()
    try:
        # Execute the notebook
        ep.preprocess(nb, {'metadata': {'path': './'}})
    except:
        # Execution failed, print a message then raise.
        msg = 'Error executing the notebook "%s".\n\n' % notebook_name
        msg += 'See notebook "%s" for the traceback.' % nb_name_output
        print(msg)
        raise
    else:
        # On successful execution, add timestamping cell
        duration = time.time() - start_time
        timestamp_cell = timestamp_cell % (time.ctime(start_time), duration)
        nb['cells'].insert(0, nbformat.v4.new_markdown_cell(timestamp_cell))
    finally:
        # Save the notebook even when it raises an error
        nbformat.write(nb, nb_name_output)
        print('* Output: ', nb_name_output)

if __name__ == '__main__':
    from pathlib import Path
    import sys

    path = '.'
    if len(sys.argv) > 1:
        path = sys.argv[1]
        assert os.path.isdir(path), 'Folder "%s" not found.' % path
    out_path = path
    if len(sys.argv) > 2:
        out_path = sys.argv[2]
        assert os.path.isdir(out_path), 'Folder "%s" not found.' % out_path
    print('Executing notebooks in "%s"... ' % os.path.abspath(path))
    pathlist = list(Path(path).glob('*.ipynb'))
    for nbpath in pathlist:
        if not (nbpath.stem.endswith('-out') or nbpath.stem.startswith('_')):
            print()
            run_notebook(nbpath, out_path=out_path)



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014-2015 Antonino Ingargiola <tritemio@gmail.com>
#
"""
This module contains routines to load and preprocess timestamp data
saved by the NI board (multi-ch data) using two 32bit words (timestamp
and detector) for each photon.

This data is produced by the 4 and 8-spot smFRET setup, first generation.
"""

from __future__ import absolute_import, division
from builtins import range, zip

import os
import numpy as np

from ..utils.misc import pprint

## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#  DATA LOADING
#
def read_int32_int32_file(fname, n_bytes_to_read=-1):
    """Read the data file with 32+32 bit format (int32 version)."""
    try: f = open(fname, 'rb')
    except IOError:
        fname += '.dat'
        f = open(fname, 'rb')

    ## Reading the header
    l1 = f.readline(); l2 = f.readline(); l3 = f.readline()
    words_per_photon = l2.split()[-1]
    assert words_per_photon == b'2'

    ##  Reading data in int32
    bytes_in_file = os.path.getsize(fname) - f.tell()
    if n_bytes_to_read < 4:
        n_bytes_to_read = bytes_in_file
    N_bytes = (int(min(n_bytes_to_read, bytes_in_file))//4)*4
    data = np.ndarray(shape=(N_bytes/4,), dtype='>i4', buffer=f.read(N_bytes))
    detector = data[::2]+1
    ph_times = (data[1::2]-data[1])
    assert ((detector < 17)*(detector >= 0)).all()
    return ph_times, detector.astype('uint8')

## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#  DATA CONVERSION
#

def swap_donor_acceptor(detectors, nch=4):
    """Swap the donor and the acceptor channels."""
    donors = detectors > nch
    acceptors = -donors
    det_d = detectors[donors]
    det_a = detectors[acceptors]
    detectors[donors] = det_d - nch
    detectors[acceptors] = det_a + nch
    return detectors

def load_data_ordered16(fname, n_bytes_to_read=-1, nch=8, swap_D_A=False,
                        remap_D=False, remap_A=False, mute=False):
    """Load data, unroll the 32bit overflow and order in increasing order."""
    pprint(' - Loading data "%s" ... ' % fname, mute)
    ph_times, detector = read_int32_int32_file(fname, n_bytes_to_read)
    pprint(" [DONE]\n", mute)
    pprint(" - Processing data ... ", mute)
    if remap_D:
        pprint("\n   - Inverting DONOR pixels order ... ", mute)
        detector[detector <= nch] = nch+1-detector[detector <= nch]
        pprint(" [DONE]\n", mute)
    if remap_A:
        pprint("\n   - Inverting ACCEPTOR pixels order ... ", mute)
        detector[detector > nch] = 2*nch+1-(detector[detector > nch] - nch)
        pprint(" [DONE]\n", mute)
    if swap_D_A:
        pprint("\n   - Swapping D and A channels ... ", mute)
        detector = swap_donor_acceptor(detector, nch=8)
        pprint(" [DONE]\n", mute)
    ph_times_m, red, ph_times_ma = unwind_uni(ph_times, detector)
    pprint("   [DONE Processing]\n", mute)

    return ph_times_m, red, ph_times_ma

def unwind_uni(times, det, nch=8, times_nbit=28, debug=True):
    """64bit conversion and merging of corresponding D/A channels."""
    diff = lambda a: a[1:]-a[:-1]
    ts_max = 2**times_nbit
    num_spad = nch*2

    times_ma = [times[det==d].astype('int64') for d in range(1, num_spad+1)]
    for i, t in enumerate(times_ma):
        ## This debug check is a tautology (cumsum of a bool mask >= 0)
        #if debug: assert (cumsum((diff(t) < 0)) >= 0).all()
        t[1:] += np.cumsum((diff(t) < 0), dtype='int64')*ts_max
        if debug: assert (np.diff(t) > 0).all()
        #print i, (t < 0).sum(), (diff(t) < 0).sum(), "\n"

    ph_times_m, red = nch*[0], nch*[0]
    bones = lambda n: np.ones(n, dtype=bool)
    bzeros = lambda n: np.zeros(n, dtype=bool)
    for i in range(nch):
        # Merge channels and sort
        ph_times_m[i] = np.hstack([times_ma[i], times_ma[i+nch]])
        index_sort = ph_times_m[i].argsort()
        red[i] = np.hstack([bzeros(times_ma[i].size),
                            bones(times_ma[i+nch].size)])
        red[i] = red[i][index_sort]
        ph_times_m[i] = ph_times_m[i][index_sort]
    return ph_times_m, red, times_ma


def unwind_uni_c(times, det, nch=8, times_nbit=28, debug=True):
    """64bit conversion and merging of corresponding D/A channels.
    This version is identical to the cython version but uses numpy sort.
    """
    cumsum, hstack, int16, int64 = np.cumsum, np.hstack, np.int16, np.int64
    diff = lambda a: a[1:]-a[:-1]
    bones = lambda n: np.ones(n, dtype=bool)
    bzeros = lambda n: np.zeros(n, dtype=bool)

    num_spad = nch*2
    ts_max = 2**times_nbit
    ph_times_m, A_det = [[]]*nch, [[]]*nch
    for ich in range(nch):
        det_d, det_a = ich+1, ich+1+nch
        t_d1, t_a1 = times[det == det_d], times[det == det_a]
        t_d = t_d1.astype(int64)
        t_d += hstack([0, cumsum((diff(t_d1)<0), dtype=int16)])*ts_max
        t_a = t_a1.astype(int64)
        t_a += hstack([0, cumsum((diff(t_a1)<0), dtype=int16)])*ts_max
        del t_d1, t_a1

        T = hstack([t_d, t_a])
        index_sort = T.argsort(kind='mergesort')

        ph_times_m[ich] = T[index_sort]
        A_det[ich] = hstack([bzeros(t_d.size), bones(t_a.size)])[index_sort]
        del index_sort, T
    return ph_times_m, A_det

def unwind_uni_o(times, det, nch=8, times_nbit=28, debug=True):
    """64bit conversion and merging of corresponding D/A channels.
    This version is about 10% faster than unwind_uni().
    """
    diff = lambda a: a[1:]-a[:-1]
    bones = lambda n: np.ones(n, dtype=bool)
    bzeros = lambda n: np.zeros(n, dtype=bool)

    num_spad = nch*2
    expo = (2**times_nbit)
    ph_times_m, A_det = [[]]*nch, [[]]*nch
    for ich in range(nch):
        det_d, det_a = ich+1, ich+1+nch
        #a d_mask, a_mask = (det == det_d), (det == det_a)

        #t_d, t_a = times[d_mask].astype(int64), times[a_mask].astype(int64)
        #t_d[1:] += cumsum((diff(t_d) < 0), dtype=int64)*expo
        #t_a[1:] += cumsum((diff(t_a) < 0), dtype=int64)*expo

        #a t_d, t_a = times[d_mask], times[a_mask]
        t_d, t_a = times[det == det_d], times[det == det_a]
        # overflows must be < 2^16
        step_d = np.cumsum((diff(t_d) < 0), dtype='int16')
        step_a = np.cumsum((diff(t_a) < 0), dtype='int16')

        t1 = np.hstack([t_d, t_a]).astype('int64')
        t2 = np.hstack([0, step_d, 0, step_a])
        #ph_times_m[ich] = NE.evaluate("t1+t2*expo") # this is slower!
        ph_times_m[ich] = t1 + t2*expo
        del t1, t2
        #ph_times_m[ich] = concatenate([t_d,t_a]).astype(int64) + \
        #        hstack([0,step_d,0,step_a])*expo
        #ph_times_m[ich] = concatenate([t_d, t_a])
        index_sort = ph_times_m[ich].argsort(kind='mergesort')
        ph_times_m[ich] = ph_times_m[ich][index_sort]
        A_det[ich] = np.hstack([bzeros(t_d.size), bones(t_a.size)])[index_sort]
        del index_sort
    return ph_times_m, A_det


if __name__ == '__main__':
    ## Some test

    #fname = gui_fname()
    #ph_times, det = read_int32_int32_file(fname)

    ph_m, a_em, t_ma = unwind_uni(ph_times, det)    # plain
    ph_mc, a_emc = unwind_uni_c(ph_times, det)      # 15% faster than plain
    ph_mo, a_emo = unwind_uni_o(ph_times, det)      # 15% faster than plain

    print([(ph == phc).all() for ph, phc in zip(ph_m, ph_mc)])
    print([(ae == aec).all() for ae, aec in zip(a_em, a_emc)])

    ## NOTE: write a compare function that takes into account same timestamps
    ##       in donor and acceptor ch.




#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2013-2016 The Regents of the University of California,
#               Antonino Ingargiola <tritemio@gmail.com>
#
"""
Routines to compute the background from an array of timestamps. This module
is normally imported as `bg` when fretbursts is imported.

The important functions are :func:`exp_fit` and :func:`exp_cdf_fit` that
provide two (fast) algorithms to estimate the background without binning.
These functions are not usually called directly but passed to
:meth:`Data.calc_bg` to compute the background of a measurement.

See also :func:`exp_hist_fit` for background estimation using an histogram fit.
"""

from __future__ import absolute_import
from builtins import zip, range

import numpy as np
from .ph_sel import Ph_sel
from .utils.misc import pprint
from .fit import exp_fitting
from .fit.gaussian_fitting import gaussian_fit_hist


def raw_fit(ph, clk_p=12.5e-9, residuals=False, tail_min_us=None):
    """Compute the "raw" rate: (number of ph / duration).

    Note:
        `tail_min_us` argument is there only to have the same signature
        of the other background fit functions but is ignored.
    """
    Lambda = ph.size/((ph[-1]-ph[0])*clk_p)
    if residuals:
        resid = exp_fitting.get_residuals(np.diff(ph), 1./(Lambda*clk_p))
        return Lambda, np.abs(resid).max()*100
    else:
        return Lambda, 0


def _compute_error(residuals, x_residuals, error_metrics):
    assert error_metrics in ['KS', 'CM', None]
    error = None
    if error_metrics == 'KS':
        error = np.abs(residuals).max()*100
    elif error_metrics == 'CM':
        error = np.trapz(residuals**2, x=x_residuals)
    return error

def _exp_fit_generic(ph, fit_fun, tail_min_us=None, tail_min_p=0.1,
                     clk_p=12.5e-9, error_metrics=None):
    """Computes BG rates on timestamp delays above a min. value.

    Compute a background rate, selecting waiting-times (delays) larger than a
    minimum threshold.

    You need to pass the specific fitting function as `fit_fun`.

    Arguments:
        ph (array): timestamps from which estimete the background.
        fit_fun (function): function to use for background fitting.
        tail_min_us (float): minimum photon separation threshold in us for
            photons to be considered as background.
        tail_min_p (int): min threshold in percentace. ONly used when
            `tail_min_us` is None. Deprecated.
        clk_p (float): unit of timestamps in `ph` (seconds).
        error_metrics (string or None): Valid values are 'KS' or 'CM'.
            'KS' (Kolmogorov-Smirnov statistics) computes the error as the
            max of deviation of the empirical CDF from the fitted CDF.
            'CM' (Crames-von Mises) uses the L^2 distance.
            If None, no error metric is computed (returns None).

    Returns:
        2-Tuple: Estimated background rate in cps, and a "quality of fit"
        index (the lower the better) according to the chosen metric.
        If error_metrics==None, the returned "quality of fit" is None.
    """
    dph = np.diff(ph)
    if tail_min_us is None:
        tail_min = dph.max()*tail_min_p
    else:
        tail_min = tail_min_us*1e-6/clk_p

    res = fit_fun(dph, s_min=tail_min, calc_residuals=error_metrics is not None)
    Lambda, residuals, x_residuals, s_size = res

    error = _compute_error(residuals, x_residuals, error_metrics)
    Lambda /= clk_p
    return Lambda, error


def exp_fit(ph, tail_min_us=None, clk_p=12.5e-9, error_metrics=None):
    """Return a background rate using the MLE of mean waiting-times.

    Compute the background rate, selecting waiting-times (delays) larger
    than a minimum threshold.

    This function performs a Maximum Likelihood (ML) fit. For
    exponentially-distributed waiting-times this is the empirical mean.

    Arguments:
        ph (array): timestamps array from which to extract the background
        tail_min_us (float): minimum waiting-time in micro-secs
        clk_p (float): clock period for timestamps in `ph`
        error_metrics (string or None): Valid values are 'KS' or 'CM'.
            'KS' (Kolmogorov-Smirnov statistics) computes the error as the
            max of deviation of the empirical CDF from the fitted CDF.
            'CM' (Crames-von Mises) uses the L^2 distance.
            If None, no error metric is computed (returns None).

    Returns:
        2-Tuple: Estimated background rate in cps, and a "quality of fit"
        index (the lower the better) according to the chosen metric.
        If error_metrics==None, the returned "quality of fit" is None.

    See also:
        :func:`exp_cdf_fit`, :func:`exp_hist_fit`
    """
    return _exp_fit_generic(ph, fit_fun=exp_fitting.expon_fit,
                            tail_min_us=tail_min_us, clk_p=clk_p,
                            error_metrics=error_metrics)

def exp_cdf_fit(ph, tail_min_us=None, clk_p=12.5e-9, error_metrics=None):
    """Return a background rate fitting the empirical CDF of waiting-times.

    Compute the background rate, selecting waiting-times (delays) larger
    than a minimum threshold.

    This function performs a least square fit of an exponential Cumulative
    Distribution Function (CDF) to the empirical CDF of waiting-times.

    Arguments:
        ph (array): timestamps array from which to extract the background
        tail_min_us (float): minimum waiting-time in micro-secs
        clk_p (float): clock period for timestamps in `ph`
        error_metrics (string or None): Valid values are 'KS' or 'CM'.
            'KS' (Kolmogorov-Smirnov statistics) computes the error as the
            max of deviation of the empirical CDF from the fitted CDF.
            'CM' (Crames-von Mises) uses the L^2 distance.
            If None, no error metric is computed (returns None).

    Returns:
        2-Tuple: Estimated background rate in cps, and a "quality of fit"
        index (the lower the better) according to the chosen metric.
        If error_metrics==None, the returned "quality of fit" is None.

    See also:
        :func:`exp_fit`, :func:`exp_hist_fit`
    """
    return _exp_fit_generic(ph, fit_fun=exp_fitting.expon_fit_cdf,
                            tail_min_us=tail_min_us, clk_p=clk_p,
                            error_metrics=error_metrics)


def exp_hist_fit(ph, tail_min_us, binw=50e-6, clk_p=12.5e-9,
                 weights='hist_counts', error_metrics=None):
    """Compute background rate with WLS histogram fit of waiting-times.

    Compute the background rate, selecting waiting-times (delays) larger
    than a minimum threshold.

    This function performs a Weighed Least Squares (WLS) fit of the
    histogram of waiting times to an exponential decay.

    Arguments:
        ph (array): timestamps array from which to extract the background
        tail_min_us (float): minimum waiting-time in micro-secs
        binw (float): bin width for waiting times, in seconds.
        clk_p (float): clock period for timestamps in `ph`
        weights (None or string): if None no weights is applied.
            if is 'hist_counts', each bin has a weight equal to its counts
            if is 'inv_hist_counts', the weight is the inverse of the counts.
        error_metrics (string or None): Valid values are 'KS' or 'CM'.
            'KS' (Kolmogorov-Smirnov statistics) computes the error as the
            max of deviation of the empirical CDF from the fitted CDF.
            'CM' (Crames-von Mises) uses the L^2 distance.
            If None, no error metric is computed (returns None).

    Returns:
        2-Tuple: Estimated background rate in cps, and a "quality of fit"
        index (the lower the better) according to the chosen metric.
        If error_metrics==None, the returned "quality of fit" is None.

    See also:
        :func:`exp_fit`, :func:`exp_cdf_fit`
    """
    assert np.size(ph) > 0
    dph = np.diff(ph)
    tail_min = tail_min_us*1e-6/clk_p
    binw_clk = binw/clk_p
    bins = np.arange(0, dph.max() - tail_min + 1, binw_clk)

    res = exp_fitting.expon_fit_hist(dph, bins=bins, s_min=tail_min,
                                     weights=weights,
                                     calc_residuals=error_metrics is not None)

    Lambda, residuals, x_residuals, s_size = res
    error = _compute_error(residuals, x_residuals, error_metrics)
    Lambda /= clk_p
    return Lambda, error

##
# Fit background as function of th
#
def fit_varying_min_delta_ph(d, min_delta_ph_list, bg_fit_fun=exp_fit,
                             ph_sel=Ph_sel('all'), **kwargs):
    """
    Fit the background as a function of the min photon interval threshold.

    The background is fitted for all the channels, all the periods and all the
    values in `min_ph_iterval_list`.

    Parameters
        d (Data object): the Data object containing the timestamps. You need to
            call d.calc_bg() before calling this function in order to create
            the background periods.
        min_delta_ph_list (list or array): list of minimum photon separation
            values above which photons are considered background. Unit: us.
        bg_fit_fun (function): function used to fit the background.
        ph_sel (Ph_sel object): photon selection on which the background is
            computed. See :class:`fretbursts.ph_sel.Ph_sel` for details.

    Returns
        Two arrays for background rate and fit-error of shape
        (nch, nperiods, len(min_delta_ph_list)).
    """

    BG = np.zeros((d.nch, d.nperiods, np.size(min_delta_ph_list)))
    BG_err = np.zeros_like(BG)
    BG[:], BG_err[:] = None, None

    for ich in range(d.nch):
        for period, ph in enumerate(
                d.iter_ph_times_period(ich=ich, ph_sel=ph_sel)):
            for i_min, min_delta_ph in enumerate(min_delta_ph_list):
                try:
                    BG[ich, period, i_min], BG_err[ich, period, i_min] = \
                            bg_fit_fun(ph, tail_min_us=min_delta_ph,
                                       clk_p=d.clk_p, **kwargs)
                except AssertionError:
                    # There are not enough delays with current threshold
                    break   # Skip remaining values in min_delta_ph_list
    return BG, BG_err


def fit_var_tail_us(d, Tail_min_us_list, t_max_s,
                    bg_fit_fun=exp_fit, ph_sel=Ph_sel('all'), **kwargs):
    """
    Fit BG of a ph_sel on all CH for all the values in `Tail_min_us_list`.

    The BG is fitted from the first `t_max_s` seconds of the measurement.

    Returns
        Two arrays for background rate and fit-error of shape
        (nch, len(min_delta_ph_list)).
    """
    assert ph_sel in [Ph_sel('all'), Ph_sel(Dex='Dem'), Ph_sel(Dex='Aem')]
    BG = np.zeros((d.nch, np.size(Tail_min_us_list)))
    BG_err = np.zeros((d.nch, np.size(Tail_min_us_list)))
    t_max_clk = t_max_s/d.clk_p

    ph_times_m_slice, A_em_slice = [], []
    for ph, a_em in zip(d.ph_times_m, d.A_em):
        ph_times_m_slice.append(ph[ph < t_max_clk])
        A_em_slice.append(a_em[ph < t_max_clk])

    Ph_times = []
    for ph, a_em in zip(ph_times_m_slice, A_em_slice):
        if ph_sel == Ph_sel(Dex='Dem'):
            Ph_times.append(ph[-a_em])
        elif ph_sel == Ph_sel(Dex='Aem'):
            Ph_times.append(ph[a_em])
        else:
            Ph_times.append(ph)

    for ch, ph_t in enumerate(Ph_times):
        for it, t in enumerate(Tail_min_us_list):
            try:
                BG[ch, it], BG_err[ch, it] = bg_fit_fun(
                    ph_t, tail_min_us=t, clk_p=d.clk_p, **kwargs)
            except:
                break # Skip remaining Tail_min
    return BG, BG_err


##
# Other functions
#
def histo(ph, bin_ms=10., t_max_s=None, clk_p=12.5e-9):
    """Returns an histogram and bins-centers of ph (ph arrival times)."""
    if t_max_s is not None:
        ph = ph[ph <= t_max_s/clk_p]
    bins = np.arange(ph[0], ph[-1]+1, (bin_ms*1e-3)/clk_p)
    H = np.histogram(ph, bins=bins)
    tt = H[0]
    ti = H[1][:-1]+0.5*(H[1][1]-H[1][0])
    return tt, ti

def gauss_fit(ph, bin_ms=10, clk_p=12.5e-9):
    """Returns the BG rate of ph calculated from the hist (PDF) of timetrace.
    """
    assert np.size(ph) > 0
    tt, ti = histo(ph=ph, bin_ms=bin_ms, clk_p=clk_p)
    #mu, sig = gaussian_fit(tt, mu_sigma_guess=[tt.mean(), tt.std()])
    mu, sig = gaussian_fit_hist(tt, mu0=tt.mean(), sigma0=tt.std())
    mask = (tt < (mu+3*sig))*(tt > (mu-3*sig))
    tt2 = tt[mask]
    #mu2, sig2 = gaussian_fit(tt2, mu_sigma_guess=[mu,sig])
    mu2, sig2 = gaussian_fit_hist(tt2, mu0=mu, sigma0=sig)
    return mu2/(bin_ms*1e-3)#, sig2/(bin_ms*1e-3)


##
# Experimental functions
#
def smart_bg(d, ich=0, bin_=50e-3, step=1):
    """BG calculation through binning (WARNING: very slow!)."""
    bg = []
    t = d.ph_times_m[ich]*d.clk_p
    t_max = np.floor(t.max())
    pprint(" Calculation started:")
    for s in np.arange(step, t_max, step):
        #if (s % (t_max/50) == 0): pprint(" %d %%" % (s/t_max*100))
        h = np.histogram(t[(t < s)*(t > (s-step))],
                         bins=np.arange(s-step, s+1e-3, bin_))
        #print(h[0])
        bg.append(h[0].min())
    pprint('\n')
    return np.array(bg)/bin_
